/*

  Simple C-like dynamic vectors/matrices

  use doble, no typedef no template

*/

#include <cstddef>
#include <iostream>

// if C++ < C++11 define nullptr
#if __cplusplus <= 199711L
  #include <cstdlib>
  #ifndef nullptr
    #include <cstddef>
    #define nullptr NULL
  #endif
#endif

using namespace std ;

// define a vector as a struct with the size of the vector
// and a pointer to its values

class Vec {
  int      _size ;
  double * values ;

public:
  
  Vec() ; // default constructor
  Vec( int _size ) ; // another constructor

  ~Vec() ; // destructor
  
  int size() const { return _size ; }
  
  // methods
  void allocate( int size ) ;
  void fill( double val ) ;
  void fill_random() ;

  // magic methods, redefine operators!
  double
  operator () // name of the operator
  ( int i ) const // const is used to inform the compiler that operator do not modifty contents
  { return values[i] ; }   

  double & // return the reference (i.e. the address) of values[i]
  operator () ( int i )
  { return values[i] ; }   

} ;

class Mat {
  int       _nrow, _ncol ;
  double  * memory_block ; // allocated memory for the matrix
  double ** values ; // pointer to the rows of the matrix

public:

  Mat() ; // default constructor
  Mat( int _nr, int _nc ) ; // another constructor

  ~Mat() ; // destructor

  int ncol() const { return _ncol ; }
  int nrow() const { return _nrow ; }
  
  void allocate( int nrow, int ncol ) ;
  void fill_random( ) ;
  void fill( double val ) ;

  // magic methods, redefine operators!
  double
  operator () // name of the operator
  ( int i, int j ) const ; // const is used to inform the compiler that operator do not modifty contents

  double & // return the reference (i.e. the address) of values[i]
  operator () ( int i, int j ) ;

} ;

// magic methods, redefine operators!
inline
double
Mat::operator () // name of the operator
( int i, int j ) const // const is used to inform the compiler that operator do not modifty contents
{ return values[i][j] ; }   

inline
double & // return the reference (i.e. the address) of values[i]
Mat::operator () ( int i, int j )
{ return values[i][j] ; }

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

ostream &
operator << ( ostream & stream, Vec const & vec ) ;

ostream &
operator << ( ostream & stream, Mat const & mat ) ;


void
Mat_x_Vec( struct Mat const & mat, // ensure the compiler that mat is not modified inside...
           struct Vec const & vec,
           struct Vec       & res ) ;

void
Mat_x_Mat_123( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;
void
Mat_x_Mat_132( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;
void
Mat_x_Mat_213( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;
void
Mat_x_Mat_231( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;
void
Mat_x_Mat_312( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;
void
Mat_x_Mat_321( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) ;

void
Mat_x_Mat_with_tiling( struct Mat const & A, 
                       struct Mat const & B,
                       struct Mat       & C,
                       int tsize ) ;
