#include "mv.hh"

#include <iostream>
#include <stdlib.h> // for rand() function

using namespace std ;

// initialize structure to a matrix of size 0 x 0
Mat::Mat( ) {
  cout << "Call constructor for Mat\n" ;
  _nrow        = 0 ;
  _ncol        = 0 ;
  memory_block = nullptr ;
  values       = nullptr ;
}

Mat::Mat( int _nr, int _nc ) : _nrow(_nr), _ncol(_nc) {
  cout << "Call constructor for Mat with size = "
       << _nr << ", " << _nc << "\n" ;

  memory_block = new double [_nrow*_ncol] ;
  values       = new double* [_nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < _nrow ; ++i )
    values[i] = memory_block + i * _ncol ;
}

Mat::~Mat() {
  cout << "Call destructor for Mat\n" ;
  // check if alreay de-allocated
  if ( memory_block != nullptr ) delete [] memory_block ;
  if ( values       != nullptr ) delete [] values ;
}

// alocate a matrix to the size: nrow x ncol
void
Mat::allocate( int nrow, int ncol ) { // USE this to distinguish arguments
  // (*this).nrow ==> this -> nrow
  if ( _nrow == nrow && _ncol == ncol ) return ; // nothing to do!

  _nrow = nrow ;
  _ncol = ncol ;

  if ( memory_block != nullptr ) delete [] memory_block ;
  memory_block = new double [nrow*ncol] ;

  if ( values != nullptr ) delete [] values ;
  values = new double* [_nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < _nrow ; ++i )
    values[i] = memory_block + i * _ncol ;
}

void
Mat::fill_random() {
  for ( int i = 0 ; i < _nrow ; ++i )
    for ( int j = 0 ; j < _ncol ; ++j )
      values[i][j] = (rand() - rand())/double(RAND_MAX) ;
}

void
Mat::fill( double val ) {
  for ( int i = 0 ; i < _nrow ; ++i )
    for ( int j = 0 ; j < _ncol ; ++j )
      values[i][j] = val ;
}

void
Mat_x_Vec( struct Mat const & mat, // ensure the compiler that mat is not modified inside...
           struct Vec const & vec,
           struct Vec       & res ) {
  // check matrix vector compatibility
  if ( mat.ncol() != vec.size() ) {
    cerr << "Mat_x_Vec, incompatible dimensions\n" ;
    exit(0) ;
  }   
  if ( res.size() != mat.nrow() ) res.allocate( mat.nrow() ) ;
  for ( int i = 0 ; i < mat.nrow() ; ++i ) {
    res(i) = 0 ;
    for ( int j = 0 ; j < mat.ncol() ; ++i )
      res(i) += mat(i,j)*vec(j) ;
  }
}

void
Mat_x_Mat_123( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_123, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
    for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
      for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_132( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_132, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
    for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
      for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_213( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_213, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
    for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
      for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_231( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_231, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
    for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
      for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_312( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_312, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
    for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
      for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_321( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol() ; ++k ) // 3
    for ( int j = 0 ; j < B.ncol() ; ++j ) // 2
      for ( int i = 0 ; i < A.nrow() ; ++i ) // 1
        C(i,j) += A(i,k)*B(k,j) ;
}

void
Mat_x_Mat_with_tiling( struct Mat const & A, 
                       struct Mat const & B,
                       struct Mat       & C,
                       int tsize ) {
  // check compatibility
  if ( A.ncol() != B.nrow() ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  C.allocate( A.nrow(), B.ncol() ) ;

  // set to 0
  C.fill( 0 ) ; // 0 int --> convertex 0 double
  
  for ( int ii = 0 ; ii < A.nrow() ; ii += tsize ) { // 1
    int ii_max = std::min(ii + tsize, A.nrow()) ;
    for ( int kk = 0 ; kk < A.ncol() ; kk += tsize ) { // 3
      int kk_max = std::min(kk + tsize, A.ncol()) ;
      for ( int jj = 0 ; jj < B.ncol() ; jj += tsize ) { // 2
        int jj_max = std::min(jj + tsize, B.ncol()) ;
        for ( int i = ii ; i < ii_max ; ++i ) // 1
          for ( int k = kk ; k < kk_max ; ++k ) // 3
            for ( int j = jj ; j < jj_max ; ++j ) // 2
              C(i,j) += A(i,k)*B(k,j) ;
      }
    }
  }
}

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

ostream &
operator << ( ostream & stream, Mat const & mat ) {
  for ( int i = 0 ; i < mat.nrow() ; ++i ) {
    stream << mat(i,0) ;
    for ( int j = 1 ; j < mat.ncol() ; ++j )
      stream << " " << mat(i,j) ;
    stream << '\n' ;
  }
  return stream ;
}
